module systems/file_system

open std/ord
open std/seq 

sig Name {}

sig Path {
  path: Seq[Name]
}

sig Node {}

disj sig File extends Node {}

// Directories aren't explicitly named... instead, there 
// are paths made up of names which may point to them
disj sig Filesystem extends Node {
  contents: Path ? -> ? Node
  } {
  // there is an empty path that maps directories to themselves
  some p: Path | EmptyPath(p) && this = p.contents 
}

fact CanonicalPath {all a, b: Path| a.path = b.path => a = b}

fact CanonicalSequence {all a, b: Seq[Name] | a..SeqEquals(b) => a = b}

fact CanonicalNode {all a, b: Node | EquivNode(a, b) => a = b}

fact FileDirPartition {
  File + Filesystem = Node
}

// "Tree structure" not a statement about being acyclic...
// When you hit a directory on your path from your entries, that 
// directory has entries consistent with yours. "Treeness" refers 
// to the fact that things rooted under you "inherit" your mappings.
// Removal of g in File... disallows existence of files
fact TreeStructure {
  all f: Filesystem | all p, q: Path | {
    let g = p.f::contents | {
      (g in File && EmptyPath(q)) => g = append(p, q).f::contents
      else (q.g::contents = append(p, q).f::contents)
    }
  }
}

// If there is a path q.x then there is a path q 
fact NoMissingPathBegin {
  all p: Path | {
    !EmptyPath(p) => { 
      some q : Path | (PPathPrefix(q, p) &&  (#PathLength(p) = #PathLength(q) + 1)) 
     }
  }
}

// Tails of paths are also paths
fact NoMissingPathEnd {
  all p: Path | {
    !EmptyPath(p) => {
      some q: Path | tail(p, q) 
    }
  }
}

/***************  UPDATE DETECTION  *****************/

// Dirty can give you clean paths
// Unused paths are dirty often
// Safely estimates updates from f to g
// Changing to iff gives exact update detection
fun dirty (f, g: Filesystem): set Path {
  all p: Path | {
    p !in result => p.f::contents = p.g::contents   
    // DirtyUpClosed
    else {all q: Path | PathPrefix(q, p) => q in result}
  }
}

/****************  RECONCILIATION  *****************/
// c and d are the synchronized versions of a and b w.r.t. dirty sets
/*fun synchronization (a, b, c, d: Node, dirtya, dirtyb: set Path) {
  all j: Path | {
    relevant(a, b, j) => {
      j !in dirtya => 
                (EquivNode(j.c::contents, j.d::contents) && 
                 EquivNode(j.c::contents, j.b::contents))
      j !in dirtyb => 
                (EquivNode(j.c::contents, j.d::contents) && 
                 EquivNode(j.c::contents, j.a::contents))
      isDirAB(a, b, j) => isDirAB(c, d, j)
      (j in dirtya && j in dirtyb && !isDirAB(a, b, j)) => 
                (EquivNode(j.c::contents, j.a::contents) && 
                 EquivNode(j.d::contents, j.b::contents))
    }
  }
} 
*/
// Synchronized filesystem after some designated path
fun afterPSynchronization (a, b, c, d: Node, dirtya, dirtyb: set Path, p: Path) {
  relevant(a, b, p)
  all pq: Path | {
    (PathPrefix(p, pq) && relevant(a, b, pq))
        => {
             pq !in dirtya => (pq.c::contents = pq.d::contents && 
                               pq.c::contents = pq.b::contents)
             pq !in dirtyb => (pq.c::contents = pq.d::contents && 
                               pq.c::contents = pq.a::contents)
             isDirAB(a, b, pq) => isDirAB(c, d, pq)
             (pq in dirtya && pq in dirtyb && !isDirAB(a, b, pq)) =>
                             (pq.c::contents = pq.a::contents && 
                              pq.d::contents = pq.b::contents)				
           }
  }
}

// A synchronization after p such that nothing else changes
fun syncp (p: Path, c, d, a, b: Node, dirtya, dirtyb: set Path) {
  afterPSynchronization(a, b, c, d, dirtya, dirtyb, p)
  all q: Path | {
    incomparable(p, q) => 
             (q.c::contents = q.a::contents && 
              q.d::contents = q.b::contents)
    (PathPrefix(q, p) && isDirAB(a, b, q)) => isDirAB(c, d, q)
  }
}

fun recon(a, b, c, d: Node, p: Path, dirtya, dirtyb: set Path) {
  all q: Path | {
    incomparable(p, q) => (q.c::contents = q.a::contents &&
                           q.d::contents = q.b::contents)
    (q !in Paths(a) && q !in Paths(b)) => 
                     (q !in Paths(c) && q !in Paths(d)) 
    (q in Paths(a) && q in Paths(b)) =>
                     (q in Paths(c) && q in Paths(d))
    isDirAB(a, b, q) => isDirAB(c, d, q)
  }
  (p !in Paths(a) && p !in Paths(b))
     => (c = a && d = b)
  else (p !in dirtya && p !in dirtyb)
     => (c = a && d = b)
  else isDirAB(a, b, p)
     => {no childrenAB(a, b, p) => (c = a && d = b) 
         else all q in childrenAB(a, b, p) | {
              reconhelper(a, b, c, d, q, dirtya, dirtyb)
         }} 
  else p !in dirtya 
     => (d = b && filesystemOverwrite(a, c, b, p))
  else p !in dirtyb
     => (c = a && filesystemOverwrite(b, d, a, p)) 
  else (c = a && d = b) 
}

fun reconhelper(a, b, c, d: Node, p: Path, dirtya, dirtyb: set Path) {
  (p !in dirtya && p !in dirtyb)
     => (p.c::contents = p.a::contents && p.d::contents = p.b::contents)
  else isDirAB(a, b, p)
     => {no childrenAB(a, b, p) => (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
         else all q in childrenAB(a, b, p) | {
              reconhelper2(a, b, c, d, q, dirtya, dirtyb)
         }} 
  else p !in dirtya 
     => (p.d::contents = p.b::contents && filesystemOverwrite(a, c, b, p))
  else p !in dirtyb
     => (p.c::contents = p.a::contents && filesystemOverwrite(b, d, a, p)) 
  else (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
}

fun reconhelper2(a, b, c, d: Node, p: Path, dirtya, dirtyb: set Path) {
  (p !in dirtya && p !in dirtyb)
     => (p.c::contents = p.a::contents && p.d::contents = p.b::contents)
  else isDirAB(a, b, p)
     => {no childrenAB(a, b, p) => (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
         else all q in childrenAB(a, b, p) | {
              reconhelper3(a, b, c, d, q, dirtya, dirtyb)
         }} 
  else p !in dirtya 
     => (p.d::contents = p.b::contents && filesystemOverwrite(a, c, b, p))
  else p !in dirtyb
     => (p.c::contents = p.a::contents && filesystemOverwrite(b, d, a, p)) 
  else (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
}

fun reconhelper3(a, b, c, d: Node, p: Path, dirtya, dirtyb: set Path) {
  (p !in dirtya && p !in dirtyb)
     => (p.c::contents = p.a::contents && p.d::contents = p.b::contents)
  else isDirAB(a, b, p)
     => {no childrenAB(a, b, p) => (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
         else all q in childrenAB(a, b, p) | {
              reconhelper4(a, b, c, d, q, dirtya, dirtyb)
         }} 
  else p !in dirtya 
     => (p.d::contents = p.b::contents && filesystemOverwrite(a, c, b, p))
  else p !in dirtyb
     => (p.c::contents = p.a::contents && filesystemOverwrite(b, d, a, p)) 
  else (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
}

fun reconhelper4(a, b, c, d: Node, p: Path, dirtya, dirtyb: set Path) {
  (p !in dirtya && p !in dirtyb)
     => (p.c::contents = p.a::contents && p.d::contents = p.b::contents)
  else isDirAB(a, b, p)
     => {no childrenAB(a, b, p) => (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
         else all q in childrenAB(a, b, p) | {
              reconhelper5(a, b, c, d, q, dirtya, dirtyb)
         }} 
  else p !in dirtya 
     => (p.d::contents = p.b::contents && filesystemOverwrite(a, c, b, p))
  else p !in dirtyb
     => (p.c::contents = p.a::contents && filesystemOverwrite(b, d, a, p)) 
  else (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
}

// Terminating
fun reconhelper5(a, b, c, d: Node, p: Path, dirtya, dirtyb: set Path) {
  (p !in dirtya && p !in dirtyb)
     => (p.c::contents = p.a::contents && p.d::contents = p.b::contents)
  else p !in dirtya 
     => (p.d::contents = p.b::contents && filesystemOverwrite(a, c, b, p))
  else p !in dirtyb
     => (p.c::contents = p.a::contents && filesystemOverwrite(b, d, a, p)) 
  else (p.c::contents = p.a::contents && p.d::contents = p.b::contents) 
}

/******************  HELPERS  ********************/

fun EmptyPath (p: Path) {
  p.path..SeqIsEmpty()
}

// a <= b
fun PathPrefix (a, b: Path) {
  b.path..SeqStartsWith(a.path)
}

// a < b
fun PPathPrefix (a, b: Path) {
  PathPrefix(a, b) && #PathLength(a) != #PathLength(b)
}

fun tail (p: Path, q: Path) {
  p.path..SeqRest() = q.path
}

// Must be used in conjunction with #
fun PathLength (p: Path): set SeqIdx  {
  result = SeqInds(p.path)
}

// All paths at some node
fun Paths(f: Filesystem): set Path {
  result = Node.~(f.contents)
}

/*  
fun LongestPath (f: Filesystem): Path {
  result in Node.~f::contents
  all q: Path - result | all m: Node | {
    q->m in f.contents => #PathLength(q) =< #PathLength(result)
  }
}
*/

// Path Append
fun append (a, b: Path): Path {
  result = {r: Path | {
    #PathLength(r) = #PathLength(a) + #PathLength(b)
    PathPrefix(a, r)
    all i: r.path..SeqInds() | {
      #OrdPrevs(i) >= #PathLength(a) =>
         some k: SeqIdx | {
           #OrdPrevs(k) + #PathLength(a)= #OrdPrevs(i)
           b.path..SeqAt(k) = r.path..SeqAt(i)
         }
    }
  }}
}

// Returns paths for children
fun children (f: Filesystem, p: Path): set Path {
   result = {q: Path | {
     q in Node.~f::contents
     PathPrefix(p, q)
     #PathLength(p) + 1  = #PathLength(q)
   }}
}

fun childrenAB (f, g: Filesystem, p: Path): set Path {
  result = (children(f, p) + children(g, p))
}

fun isDir (f: Filesystem, p: Path) {
  some (Filesystem & p.f::contents)
}

fun isDirAB (f, g: Filesystem, p:Path) {
  isDir(f, p) && isDir(g, p) 
}

// Equivalent filesystems map same paths to same files and directory structure
// On empty inputs, returns true
fun EquivNode (f, g: Node) {
  (f = g) || 
  (f in Filesystem && g in Filesystem && Paths(f) = Paths(g) &&
    all p: Path | all e: File | {
      p->e in f.contents <=> p->e in g.contents
    }
  )  
}

// Path p is relevant if all its ancestors refer to dirs in both dirs
// Note restriction that a, b are Filesystems and not just nodes... 
// Otherwise, synchronization is often trivially true
fun relevant (a, b: Filesystem, p: Path) {
  all q: Path | {
    PPathPrefix(q, p) => isDirAB(a, b, q)
  }
}

// Two paths are on "different branches"
fun incomparable (p, q: Path) {
  !PathPrefix(p, q) && !PathPrefix(q, p)
}

// Replace subtree at p in T with S's
fun filesystemOverwrite (t, s: Node, p: Path): Node {
  all q: Path | {
    PathPrefix(p, q) => q.result::contents = q.s::contents
  }
}

/***************  CHECKING ***************/
fun Simple () {
some o, a, b, c, d: Filesystem | some p: Path | some dirtya, dirtyb: set Path |{
  dirty(o, dirtya, a)
  dirty(o, dirtyb, b)
  dirtyb != Path
  recon(a, b, c, d, p, dirtya, dirtyb)
  #a.contents > 2
  o != a
  b != d
}
}

// Recon satisfies the requirements for synchronization
// Works if you exclude cases that would use recursion
assert Soundness { 
  all a, b, o: Filesystem | all dirtya, dirtyb: set Path | {
      all z: Path | all c, d: Filesystem | {
        (dirty(o, dirtya, a) && dirty(o, dirtyb, b) && 
          relevant(a, b, z) && recon(a, b, c, d, z, dirtya, dirtyb)) =>
                                   (syncp(z, c, d, a, b, dirtya, dirtyb))            
      }
  }
}

assert reconGensAll {
  all o, a, b: Filesystem | all z: Path | all dirtya, dirtyb: set Path | {
    (dirty(o, dirtya, a) && dirty(o, dirtyb, b) && some c, d: Filesystem | {syncp(z, c, d, a, b, dirtya, dirtyb)})
      => some c, d: Filesystem | {recon(a, b, c, d, z, dirtya, dirtyb)}
  }
}

// Only one sychronization of a pair of filesystems w.r.t.
// dirty predicates
assert Uniqueness {
  all a, b, o: Filesystem | all dirtya, dirtyb: set Path | {
     all p: Path | all c1, d1, c2, d2: Filesystem | {
      (dirty(o, dirtya, a) && dirty(o, dirtyb, b) && relevant(a, b, p) && 
                            afterPSynchronization(a, b, c1, d1, dirtya, dirtyb, p) && 
                            afterPSynchronization(a, b, c2, d2, dirtya, dirtyb, p)) =>
                                   (p.c1::contents = p.c2::contents && 
                                    p.d1::contents = p.d2::contents)            
     }
  }
}

// If a path isn't dirty in either updated replica then the path
// maps to equivalent nodes in both
assert NonDirtyAreSame { 
  all a, b, o: Filesystem | all p: Path | all dirtya, dirtyb: set Path | {
    (dirty(o, dirtya, a) && dirty(o, dirtyb, b) && p !in dirtya && p !in dirtyb) => 
             p.a::contents = p.b::contents
  }
}

// True as long as don't get rid of NoMissingPath fact and the restriction on 
// filesystem signature at once
assert NoCycles  {
  all f: Filesystem | no p: Path | {
    !EmptyPath(p) && p->f in f.contents
  }
}

run Simple for 4
check NoCycles for 6
check NonDirtyAreSame for 6
check Uniqueness for 4 but 5 Path
check reconGensAll for 4
check Soundness for /*5 but 8 Path*//*6 but 3 Name*//*6 but 1 Name*/5

