module systems/file_system

open std/ord
open std/seq 

sig Name {}

sig Path {
  path: Seq[Name]
}

sig Node {}

disj sig File extends Node{} 

// Directories aren't explicitly named... instead, there 
// are paths made up of names which may point to them
disj sig Filesystem extends Node {
  contents: Path ? -> ? Node
  } {
  // there is an empty path that maps directories to themselves
  some p: Path | EmptyPath(p) && this = p.contents 
}

fact CanonicalPath {all a, b: Path| a.path = b.path => a = b}

fact CanonicalSequence {all a, b: Seq[Name] | a..SeqEquals(b) => a = b}

fact CanonicalNode {all a, b: Node| EquivNode(a, b) => a = b}

fact FileDirPartition {
  File + Filesystem = Node
}

// "Tree structure" not a statement about being acyclic...
// When you hit a directory on your path from your entries, that 
// directory has entries consistent with yours. "Treeness" refers 
// to the fact that things rooted under you "inherit" your mappings.
// Removal of g in File... disallows existence of files
fact TreeStructure {
  all f: Filesystem | all p, q: Path | {
    let g = p.f::contents | {
      (g in File && EmptyPath(q)) => g = append(p, q).f::contents
      else (q.g::contents = append(p, q).f::contents)
    }
  }
}

// If there is a path q.x then there is a path q 
fact NoMissingPathBegin {
  all p: Path | {
    !EmptyPath(p) => { 
      some q : Path | (PPathPrefix(q, p) &&  (#PathLength(p) = #PathLength(q) + 1)) 
     }
  }
}

// Tails of paths are also paths
fact NoMissingPathEnd {
  all p: Path | {
    !EmptyPath(p) => {
      some q: Path | tail(p, q) 
    }
  }
}

/***************  UPDATE DETECTION  *****************/

fun dirty (f, g: Filesystem): set Path {
  all p: Path | {
    p !in result <=> p.f::contents = p.g::contents   
  }
}

fun conflict(o, a, b: Node) {
  !(samesort(a, b) || o = a || o = b)
}

/***************** Unison recon ******************/

fun relevant (a, b: Filesystem, p: Path) {
  all q: Path | {
    PPathPrefix(q, p) => isDirAB(a, b, q)
  }
}

fun PreserveLocalChanges(o, a, b, c, d: Node) {
  !samesort(o, a) => samesort(c, a)
  !samesort(o, b) => samesort(d, b)
}

fun PropogateOnlyUserChanges(o, a, b, c, d: Node) {
  !samesort(a, c) => samesort(b, c)
  !samesort(b, d) => samesort(a, d)
}

fun StopAtConflicts(o, a, b, c, d: Node) {
  conflict(o, a, b) => (c = a && d = b)
}

fun safe(o, a, b, c, d: Node) {
  all p: Path | {
    let o' = p.o::contents, a' = p.a::contents,
        b' = p.b::contents, c' = p.c::contents,
                            d' = p.d::contents | {
          PreserveLocalChanges(o', a', b', c', d')
          PropogateOnlyUserChanges(o', a', b', c', d')
          StopAtConflicts(o', a', b', c', d')
    }
  }
}

fun maximal(o, a, b, c, d: Node) {
  all c', d': Node | {
    safe(o, a, b, c', d') =>
      all p: Path | {
        samesort(p.c'::contents, p.d'::contents) => samesort(p.c::contents, p.d::contents)
      }
  }
}

// c and d are the synchronized versions of a and b w.r.t. dirty sets
fun synchronization (a, b, c, d: Node, dirtya, dirtyb: set Path) {
  all j: Path | {
    relevant(a, b, j) => {
      j !in dirtya => 
                (j.c::contents = j.d::contents && 
                 j.c::contents = j.b::contents)
      j !in dirtyb => 
                (j.c::contents = j.d::contents && 
                 j.c::contents = j.a::contents)
      isDirAB(a, b, j) => isDirAB(c, d, j)
      (j in dirtya && j in dirtyb && !isDirAB(a, b, j)) => 
                (j.c::contents = j.a::contents && 
                 j.d::contents = j.b::contents)
    }
  }
} 

// Nonconflicting updates are propogated
fun unisonRecon(o, a, b, c, d: Node) {
  all p: Path | {
    (samesort(p.o::contents, p.a::contents) && samesort(p.o::contents, p.b::contents)) =>
      (samesort(p.a::contents, p.c::contents) && samesort(p.b::contents, p.d::contents))
    else (samesort(p.o::contents, p.a::contents) &&
           {all q: Path | PathPrefix(q, p) => !conflict(q.o::contents, q.a::contents, q.b::contents)}) =>
      (samesort(p.b::contents, p.c::contents) && samesort(p.b::contents, p.d::contents))
    else (samesort(p.o::contents, p.b::contents) &&
           {all q: Path | PathPrefix(q, p) => !conflict(q.o::contents, q.a::contents, q.b::contents)}) =>
      (samesort(p.a::contents, p.c::contents) && samesort(p.a::contents, p.d::contents))
    else (samesort(p.a::contents, p.c::contents) && samesort(p.b::contents, p.d::contents))
  }
}

// Nonconflicting updates are propogated
// Is not maximal in the sense that it does not synchronize everything it should
/*fun BadUnisonRecon(o, a, b, c, d: Node) {
  all p: Path | {
    (samesort(p.o::contents, p.a::contents) && samesort(p.o::contents, p.b::contents)) =>
      (samesort(p.a::contents, p.c::contents) && samesort(p.b::contents, p.d::contents))
    else (samesort(p.o::contents, p.a::contents) &&
                                  !conflict(p.o::contents, p.a::contents, p.b::contents)) =>
      (samesort(p.b::contents, p.c::contents) && samesort(p.b::contents, p.d::contents))
    else (samesort(p.o::contents, p.b::contents) &&
                                  !conflict(p.o::contents, p.a::contents, p.b::contents)) =>
      (samesort(p.a::contents, p.c::contents) && samesort(p.a::contents, p.d::contents))
    else (samesort(p.a::contents, p.c::contents) && samesort(p.b::contents, p.d::contents))
  }
}
*/
/******************  HELPERS  ********************/

fun EmptyPath (p: Path) {
  p.path..SeqIsEmpty()
}

// a <= b
fun PathPrefix (a, b: Path) {
  b.path..SeqStartsWith(a.path)
}

// a < b
fun PPathPrefix (a, b: Path) {
  PathPrefix(a, b) && #PathLength(a) != #PathLength(b)
}

fun tail (p: Path, q: Path) {
  p.path..SeqRest() = q.path
}

// Must be used in conjunction with #
fun PathLength (p: Path): set SeqIdx  {
  result = SeqInds(p.path)
}

// All paths at some node
fun Paths(f: Filesystem): set Path {
  result = Node.~(f.contents)
}

// Path Append
fun append (a, b: Path): Path {
  result = {r: Path | {
    #PathLength(r) = #PathLength(a) + #PathLength(b)
    PathPrefix(a, r)
    all i: r.path..SeqInds() | {
      #OrdPrevs(i) >= #PathLength(a) =>
         some k: SeqIdx | {
           #OrdPrevs(k) + #PathLength(a)= #OrdPrevs(i)
           b.path..SeqAt(k) = r.path..SeqAt(i)
         }
    }
  }}
}

fun isDir (f: Filesystem, p: Path) {
  some (Filesystem & p.f::contents)
}

fun isDirAB (f, g: Filesystem, p:Path) {
  isDir(f, p) && isDir(g, p) 
}

// Equivalent filesystems map same paths to same files and directory structure
// On empty inputs, returns true
fun EquivNode (f, g: Node) {
  (f = g) || 
  (f in Filesystem && g in Filesystem && Paths(f) = Paths(g) &&
    all p: Path | all e: File | {
      p->e in f.contents <=> p->e in g.contents
    }
  )  
}

/*Not having samesort is same as an update when you look at a particular path*/
fun samesort2(a, b: Node) {
  a = b || (a in Filesystem && b in Filesystem)
}

fun samesort(a, b: Node) {
  a = b || (some (a & Filesystem) && some (b & Filesystem))
}

/*************CHECKING**********************************/

fun genExample() {
  #Filesystem = 4
  one File
  some a: Filesystem | {
    #a.contents > 2
  }
  some o, a, b, c, d: Filesystem | {
    o != a
    o != b
    a != b
    unisonRecon(o, a, b, c, d)
  }
} 

assert checkUnisonSafe {
  all o, a, b, c, d: Node | {
    unisonRecon(o, a, b, c, d) =>
     (safe(o, a, b, c, d) &&
      maximal(o, a, b, c, d))
  }
}

assert UnisonGensAll {
  all o, a, b, c, d: Filesystem | {
   (safe(o, a, b, c, d) &&
     maximal(o, a, b, c, d)) =>
       unisonRecon(o, a, b, c, d)
  }
}

assert unisonMaximal {
  all o, a, b, c, d: Node, dirtya, dirtyb: set Path | {
    (dirty(o, dirtya, a) && dirty(o, dirtyb, b) && synchronization(a, b, c, d, dirtya, dirtyb)) =>
         unisonRecon(o, a, b, c, d)
  }
}

assert laziness {
  all o, a, b: Node | safe(o, a, b, a, b)
}

assert mirroring {
  all o, a: Node | maximal(o, a, o, a, a)
}

assert briefcaseMirroring {
  all o, a: Node | unisonRecon(o, a, o, a, a)
}

assert briefcaseSymmetricMirroring {
  all o, a: Node | unisonRecon(o, o, a, a, a)
}

assert briefcase {
  all o, a, o', a', b, b', o'', a'', o''': Node | { 
    /*a is synchronized first*/
    {unisonRecon(o, a, o, a', o')
     /*b is synchronized with the updated network version*/
     unisonRecon(o, b, o', b', o'')
     /*a is synchronized again with the updated network version to get b's updates*/
     unisonRecon(o', a', o'', a'', o''')}
            /*Turns out this is the same as doing unisonRecon to begin with so*/
            /*serialized recons are the same, giving correctness of briefcase*/
         => unisonRecon(o, a, b, a'', b')
  }
}

assert maximalUnique {
  all o, a, b, c, d, c', d': Node | {
    (safe(o, a, b, c, d) && safe(o, a, b, c', d') &&
     maximal(o, a, b, c, d) && maximal(o, a, b, c', d')) =>
      all p: Path | {
        samesort(p.c::contents, p.c'::contents)
      }
  }
}

assert noConflictSuccess {
  all o, a, b: Node | {
    all p: Path | {
      !conflict(p.o::contents, p.a::contents, p.b::contents)
    } => some c: Node | {maximal(o, a, b, c, c)}
  }
}

run genExample for 5
check checkUnisonSafe for 5
check UnisonGensAll for 3
check unisonMaximal for /*5*/4
check maximalUnique for 4 
check noConflictSuccess for 4 
check laziness for 4 
check mirroring for 4
check briefcaseMirroring for 4
check briefcaseSymmetricMirroring for 4 
check briefcase for 4