module systems/file_system

open std/ord
open std/seq 

sig Name {}

sig Path {
  path: Seq[Name]
}

sig Node {}

disj sig File extends Node {}
 
// Directories aren't explicitly named... instead, there 
// are paths made up of names which may point to them
disj sig Filesystem extends Node {
  contents: Path ? -> ? Node
  } {
  // there is an empty path that maps directories to themselves
  some p: Path | EmptyPath(p) && this = p.contents 
}

// Recon maps original filesystem, two updates, and a path to new filesystems
// Dirty maps pairs of old and updated filesystems to paths 
static sig Synchronizer {
  recon: Node -> Node -> Node -> Path -> !(Node->Node), 
  dirty : (Node -> Node) -> Path 
} 

fact CanonicalPath {all a, b: Path| a.path = b.path => a = b}

fact CanonicalSequence {all a, b: Seq[Name] | a..SeqEquals(b) => a = b}

fact CanonicalNode {all a, b: Node | EquivNode(a, b) => a = b}

fact FileDirPartition {
  File + Filesystem = Node
}

// "Tree structure" not a statement about being acyclic...
// When you hit a directory on your path from your entries, that 
// directory has entries consistent with yours. "Treeness" refers 
// to the fact that things rooted under you "inherit" your mappings.
// Removal of g in File... disallows existence of files
fact TreeStructure {
  all f: Filesystem | all p, q: Path | {
    let g = p.f::contents | {
      (g in File && EmptyPath(q)) => g = append(p, q).f::contents
      else (q.g::contents = append(p, q).f::contents)
    }
  }
}

// If there is a path q.x then there is a path q 
fact NoMissingPathBegin {
  all p: Path | {
    !EmptyPath(p) => { 
      some q : Path | (PPathPrefix(q, p) &&  (#PathLength(p) = #PathLength(q) + 1)) 
     }
  }
}

// Tails of paths are also paths
fact NoMissingPathEnd {
  all p: Path | {
    !EmptyPath(p) => {
      some q: Path | tail(p, q) 
    }
  }
}

/***************  UPDATE DETECTION  *****************/

// Dirty can give you clean paths
// Unused paths are dirty often
// Safely estimates updates from f to g
// Changing to iff gives exact update detection

fact DirtyBehaviour {
  all f, g: Filesystem | all p: Path | {
    p !in Synchronizer.dirty[f][g] => p.f::contents = p.g::contents   
  }
}

fact DirtyUpClosed {
  all f, g: Filesystem | all p, q: Path | {
    (PathPrefix(p, q) && q in Synchronizer.dirty[f][g]) => 
                                  p in Synchronizer.dirty[f][g]
  }
}

/****************  RECONCILIATION  *****************/

// Synchronized filesystem after some designated path
fun afterPSynchronization (a, b, c, d: Node, dirtya, dirtyb: set Path, p: Path) {
  relevant(a, b, p)
  all pq: Path | {
    (PathPrefix(p, pq) && relevant(a, b, pq))
        => {
             pq !in dirtya => (pq.c::contents = pq.d::contents && 
                               pq.c::contents = pq.b::contents)
             pq !in dirtyb => (pq.c::contents = pq.d::contents && 
                               pq.c::contents = pq.a::contents)
             isDirAB(a, b, pq) => isDirAB(c, d, pq)
             (pq in dirtya && pq in dirtyb && !isDirAB(a, b, pq)) =>
                             (pq.c::contents = pq.a::contents && 
                              pq.d::contents = pq.b::contents)				
           }
  }
}

// A synchronization after p such that nothing else changes
fun syncp (p: Path, c, d, a, b: Node, dirtya, dirtyb: set Path) {
  afterPSynchronization(a, b, c, d, dirtya, dirtyb, p)
  all q: Path | {
    incomparable(p, q) => 
             (q.c::contents = q.a::contents && 
              q.d::contents = q.b::contents)
    (PathPrefix(q, p) && isDirAB(a, b, q)) => isDirAB(c, d, q)
  }
}

// Constraining behaviour of recon mapping
fact reconFacts {
  all o, a, b: Node | all p: Path | some abp, bap: Node |{
    let dirtya = Synchronizer.dirty[o][a], dirtyb = Synchronizer.dirty[o][b] | { 
      // p didn't change in either filesystem, so don't change anything
      (p !in dirtya && p !in dirtyb)
        => Synchronizer.recon[o][a][b][p] = a->b
      // recursive portion... recon of a dirty directory has property that it has
      // same contents as the recon of subdirectories
      else isDirAB(a, b, p) 
        => {
          no childrenAB(a, b, p) => Synchronizer.recon[o][a][b][p] = a->b
          all q: childrenAB(a, b, p) | {
            let c = Node.~(Synchronizer::recon[o][a][b][p]), 
                d = Synchronizer.recon[o][a][b][p][c],
                e = Node.~(Synchronizer.recon[o][a][b][q]), 
                f = Synchronizer.recon[o][a][b][q][e] | {
                    (children(c, p) + children(d, p)) in childrenAB(a, b, p) 
                     all s: Path | {PathPrefix(s, q) => 
                                     ((s in Paths(c) <=> s in Paths(e)) &&
                                      (s in Paths(d) <=> s in Paths(f))) }
                     (q.c::contents = q.e::contents && q.d::contents = q.f::contents)
             }
          }  
        }
      // otherwise, if only dirty in one filesystem, then overwrite the unchanged 
      else p !in dirtya 
        => filesystemOverwrite(a, abp, b, p) && Synchronizer.recon[o][a][b][p] = abp->b
      else p !in dirtyb
        => filesystemOverwrite(b, bap, a, p) && Synchronizer.recon[o][a][b][p] = a->bap
      // else it's a conflicting update so don't change anything
      else Synchronizer.recon[o][a][b][p] = a->b
    }
  }
}

/******************  HELPERS  ********************/

fun EmptyPath (p: Path) {
  p.path..SeqIsEmpty()
}

// a <= b
fun PathPrefix (a, b: Path) {
  b.path..SeqStartsWith(a.path)
}

// a < b
fun PPathPrefix (a, b: Path) {
  PathPrefix(a, b) && #PathLength(a) != #PathLength(b)
}

fun tail (p: Path, q: Path) {
  p.path..SeqRest() = q.path
}

// Must be used in conjunction with #
fun PathLength (p: Path): set SeqIdx  {
  result = SeqInds(p.path)
}

// All paths at some node
fun Paths(f: Filesystem): set Path {
  result = Node.~(f.contents)
}

/*  
fun LongestPath (f: Filesystem): Path {
  result in Node.~f::contents
  all q: Path - result | all m: Node | {
    q->m in f.contents => #PathLength(q) =< #PathLength(result)
  }
}
*/

// Path Append
fun append (a, b: Path): Path {
  result = {r: Path | {
   #PathLength(r) = #PathLength(a) + #PathLength(b)
   PathPrefix(a, r)
   all i: r.path..SeqInds() | {
     #OrdPrevs(i) >= #PathLength(a) =>
        some k: SeqIdx | {
          #OrdPrevs(k) + #PathLength(a)= #OrdPrevs(i)
          b.path..SeqAt(k) = r.path..SeqAt(i)
        }
   }
  }}
}

// Returns paths for children
fun children (f: Filesystem, p: Path): set Path {
  result = {q: Path | {
       q in Node.~f::contents
       PathPrefix(p, q)
       #PathLength(p) + 1  = #PathLength(q)
       }
  }
}

fun childrenAB (f, g: Filesystem, p: Path): set Path {
  result = (children(f, p) + children(g, p))
}

fun isDir (f: Filesystem, p: Path) {
  some (Filesystem & p.f::contents)
}

fun isDirAB (f, g: Filesystem, p:Path) {
  isDir(f, p) && isDir(g, p) 
}

// Equivalent filesystems map same paths to same files and directory structure
// On empty inputs, returns true
fun EquivNode (f, g: Node) {
  (f = g) || 
  (f in Filesystem && g in Filesystem && Paths(f) = Paths(g) &&
    all p: Path | all e: File | {
      p->e in f.contents <=> p->e in g.contents
    }
  )  
}

// Path p is relevant if all its ancestors refer to dirs in both dirs
// Note restriction that a, b are Filesystems and not just nodes... 
// Otherwise, synchronization is often trivially true
fun relevant (a, b: Filesystem, p: Path) {
  all q: Path | {
    PPathPrefix(q, p) => isDirAB(a, b, q)
  }
}

// Two paths are on "different branches"
fun incomparable (p, q: Path) {
  !PathPrefix(p, q) && !PathPrefix(q, p)
}

// Replace subtree at p in T with S's
fun filesystemOverwrite (t, s: Node, p: Path): Node {
  p !in (Node.~(s::contents) + Node.~(t::contents)) => result = t
  else all q: Path | {
     PathPrefix(p, q) => q.result::contents = q.s::contents
     else incomparable(p, q) => q.result::contents = q.t::contents
     else children(result, q) + p = children(t, q) + p
  }
}

/***************  CHECKING ***************/
fun Simple () {
  some File
}

// Recon doesn't miss cases
assert reconGensAll {
  all o, a, b: Filesystem | all z: Path | {
    let dirtya = Synchronizer.dirty[o][a], dirtyb = Synchronizer.dirty[o][b] | {
      (some dirtya && some dirtyb && some c, d: Filesystem | {syncp(z, c, d, a, b, dirtya, dirtyb)})
        => some Synchronizer.recon[o][a][b][z]
    }
  }
}

// Recon satisfies the requirements for synchronization
// Works if you exclude cases that would use recursion
assert Soundness { 
  all a, b, o: Filesystem | {
   let dirtya = Synchronizer.dirty[o][a], dirtyb = Synchronizer.dirty[o][b] | {
      all p: Path | all c, d: Filesystem | {
        (relevant(a, b, p) && Synchronizer.recon[o][a][b][p] = c->d) =>
                                   (syncp(p, c, d, a, b, dirtya, dirtyb))            
      }
    }
  }
}

// Only one sychronization of a pair of filesystems w.r.t.
// dirty predicates
assert Uniqueness {
  all a, b, o: Filesystem | {
    let dirtya = Synchronizer.dirty[o][a], dirtyb = Synchronizer.dirty[o][b] | {
     all p: Path | all c1, d1, c2, d2: Filesystem | {
      (relevant(a, b, p) && afterPSynchronization(a, b, c1, d1, dirtya, dirtyb, p) && 
                            afterPSynchronization(a, b, c2, d2, dirtya, dirtyb, p)) =>
                                   (p.c1::contents = p.c2::contents && 
                                    p.d1::contents = p.d2::contents)            
     }
   }
  }
}

// If a path is dirty in either updated replica then the path
// maps to equivalent nodes in both
assert NonDirtyAreSame { 
  all a, b, o: Filesystem | all p: Path | {
    (p !in Synchronizer.dirty[o][a] && p !in Synchronizer.dirty[o][b]) => 
             p.a::contents = p.b::contents
  }
}

// True as long as don't get rid of NoMissingPath fact and the restriction on 
// filesystem signature at once
assert NoCycles  { 
  all f: Filesystem | no p: Path | {
    !EmptyPath(p) && p->f in f.contents
  }
}

run Simple for /*3 but 4 Path*/3
check NoCycles for 3 but 4 Path
check NonDirtyAreSame for 3 but 4 Path
check Uniqueness for 3 but 4 Path
check Soundness for 3 but 4 Path
check reconGensAll for 3 but 4 Path